import os
import cv2
import numpy as np
import argparse
import sys
from pathlib import Path
from typing import Union
from tqdm import tqdm
import json
from prodict import Prodict
import copy

def fix_interactive():
    if sys.argv[0].split('/')[-1] == 'ipykernel_launcher.py':
        print("Running in interactive window...")
        sys.argv = ['']

def file_ext(name: Union[str, Path]) -> str:
    return str(name).split('.')[-1]

def is_image_ext(fname: Union[str, Path]) -> bool:
    ext = file_ext(fname).lower()
    return f'.{ext}' in [".png", ".jpeg", ".jpg"]

def is_npy_ext(fname: Union[str, Path]) -> bool:
    ext = file_ext(fname).lower()
    return f'.{ext}' in [".npy"]

def frame_to_opencv(frame):
    new_frame = copy.deepcopy(frame)
    extrinsics = np.array(new_frame["transform_matrix"])
    extrinsics_opencv = open_cv_to_open_gl(extrinsics).tolist()
    new_frame["transform_matrix"] = extrinsics_opencv
    return new_frame

def open_cv_to_open_gl(cam_matrix):
    '''
    Transform camera transformation matrix axis.
    '''
    reverse = np.diag([1, -1, -1, 1])
    cam_matrix =  cam_matrix @ reverse
    return cam_matrix    

def get_intrinsics(target_resolution):
    # Computing objaverse intrinsics
    h, w = 512, 512
    cx, cy = w / 2, h / 2
    camera_angle_x = 0.857
    f = 1 / np.tan(camera_angle_x / 2) * (w / 2)
    fl_x, fl_y = f, f

    # Rescale to target resolution
    scale_factor = target_resolution / 512
    w = int(w * scale_factor)
    h = int(h * scale_factor)
    fl_x = fl_x * scale_factor 
    fl_y = fl_y * scale_factor
    cx = cx * scale_factor
    cy = cy * scale_factor

    return fl_x, fl_y, cx, cy, w, h

def get_calibration_matrix_K_from_blender(target_resolution):
    f_in_mm = 35
    resolution_x_in_px = 512
    resolution_y_in_px = 512
    scale = target_resolution / resolution_x_in_px
    sensor_width_in_mm = 32
    sensor_height_in_mm = 32
    # the sensor height is fixed (sensor fit is horizontal), 
    # the sensor width is effectively changed with the pixel aspect ratio
    pixel_aspect_ratio = 1
    s_u = resolution_x_in_px * scale / sensor_width_in_mm / pixel_aspect_ratio 
    s_v = resolution_y_in_px * scale / sensor_height_in_mm
    

    # Parameters of intrinsic calibration matrix K
    alpha_u = f_in_mm * s_u
    alpha_v = f_in_mm * s_v
    u_0 = resolution_x_in_px * scale / 2
    v_0 = resolution_y_in_px * scale / 2

    # skew = 0 # only use rectangular pixels
    # K = Matrix(
    #     ((alpha_u, skew,    u_0),
    #     (    0  , alpha_v, v_0),
    #     (    0  , 0,        1 )))
    return alpha_u, alpha_v, u_0, v_0, scale*resolution_x_in_px, scale*resolution_y_in_px

def cam_inv(camera_matrix):
    """Function to efficiently invert between cam2world and world2cam"""
    assert camera_matrix.shape == (4, 4)
    inverse_camera_matrix = np.eye(4)
    
    R = camera_matrix[:3, :3]
    t = camera_matrix[:3, 3]
    inverse_camera_matrix[:3, :3] = R.T
    inverse_camera_matrix[:3, 3] = -R.T @ t
    return inverse_camera_matrix

def get_altitude(extrinsics):
    return extrinsics[2, 3]
    
if __name__ == '__main__':
    fix_interactive()

    DATA_DIR = Path("path/to/data/dir")
    
    parser = argparse.ArgumentParser()
    parser.add_argument("--dataset", type=str, default="objaverse/small")
    parser.add_argument("--raw_data_name", type=str, default="raw")
    parser.add_argument("--processed_data_name", type=str, default="square128")
    parser.add_argument("--target_resolution", type=int, default=128)
    parser.add_argument("--scale_camera", type=bool, default=True)
    parser.add_argument("--above", action='store_true')
    args = parser.parse_args()

    objaverse_data = DATA_DIR / args.dataset
    raw_data = Path(args.raw_data_name)
    processed_data = Path(args.processed_data_name)
    target_resolution = args.target_resolution

    raw_scenes = os.listdir(objaverse_data / raw_data)


    for raw_scene in tqdm(raw_scenes, desc='Preprocessing scenes...'):
        scene_path = objaverse_data / raw_data / raw_scene
        processed_scene_path = objaverse_data / processed_data / raw_scene
        transforms = {
            "camera_model": "OPENCV",
            "frames": []
        }
        transforms_opencv = copy.deepcopy(transforms)

        if not os.path.exists(processed_scene_path):
            os.makedirs(processed_scene_path / "rgb")
        
        raw_images = [k for k in os.listdir(scene_path) if is_image_ext(k)]
        metadata = [k for k in os.listdir(scene_path) if is_npy_ext(k)]

        raw_images.sort()
        metadata.sort()

        for k in range(len(raw_images)):

            # 1. Process camera matrices adn transform.json
            data = metadata[k]
            frame = {}
            frame["file_path"] = f"rgb/{data.split('.')[0]}.png"

            data_path = scene_path / data
            processed_data_path = processed_scene_path

            extrinsics = np.load(data_path)
            extrinsics = np.vstack([extrinsics, [0, 0, 0, 1]])

            c_sn = 1.3 # Camera distance in shapenet
            f_sn = 1.02 # Focal length in shapenet
            f_obj = 1.09375 # Default focal length in objaverse
            c_obj = 2 # Default camera distance in objaverse
            if args.scale_camera:
                # This scales the camera so the object is of same size as in the shapenet dataset
                extrinsics[0:3, 3] = extrinsics[0:3, 3] * (c_sn / f_sn) * (f_obj / c_obj)



            extrinsics = cam_inv(extrinsics)
            if args.above:
                altitude = get_altitude(extrinsics)
                if altitude < 0:
                    continue # go to next frame
            frame["transform_matrix"] = extrinsics.tolist()

            # fl_x, fl_y, cx, cy, w, h = get_intrinsics(target_resolution)
            fl_x, fl_y, cx, cy, w, h = get_calibration_matrix_K_from_blender(target_resolution)
            frame["fl_x"] = fl_x
            frame["fl_y"] = fl_y
            frame["cx"] = cx
            frame["cy"] = cy
            frame["w"] = w
            frame["h"] = h

            transforms["frames"].append(frame)
            transforms_opencv["frames"].append(frame_to_opencv(frame))

            #2. Resize and write images
            raw_image = raw_images[k]
            image_path = scene_path / raw_image
            processed_image_path = processed_scene_path / "rgb" / raw_image
            image = cv2.imread(str(image_path), cv2.IMREAD_UNCHANGED)
            #make mask of where the transparent bits are
            trans_mask = image[:,:,3] == 0
            #replace areas of transparency with white and not transparent
            image[trans_mask] = [255, 255, 255, 255]
            image = cv2.resize(image, (target_resolution, target_resolution), interpolation=cv2.INTER_AREA)

            cv2.imwrite(str(processed_image_path), image)



        # save json 
        with open(processed_scene_path / "transforms.json", 'w') as f:
            json.dump(transforms, f, indent=4)
        with open(processed_scene_path / "transforms_opencv.json", 'w') as f:
            json.dump(transforms_opencv, f, indent=4)
