import os
import PIL.Image
from prodict import Prodict
import numpy as np
import matplotlib.pyplot as plt
from pathlib import Path
import json
import random
import torchvision.transforms as tr
import torch
import tqdm
import argparse
import sys

if __name__ == '__main__':
    sys.path.append(str(Path(__file__).resolve().parents[3]))

from datasets.dataset import DataReader


def normalize(img_arr):
    return img_arr * 0.5 + 0.5

def sample_views(idx_obj, n_views=10):
    "Returns a array of views of a scene specified by idx_obj. Retunred array has shape (n_views, H, W, C) with values in [0, 1]"
    views = []
    for i in range(n_views):
        img_arr = mv_dset.get_image(idx_obj, i)['img'].numpy()
        views.append(normalize(img_arr).transpose(1, 2, 0))
    return np.stack(views, axis=0)

def get_white_pixels_mask(arr_img):
    """args:
        arr_img: (..., C) with values in [0, 1]
    returns: 
        mask (...) with values in {0, 1}. 1 if pixel is white, 0 otherwise
    """
    return np.all(arr_img == 1, axis=-1)

def expand_mask(mask, num_channels=3):
    "expands a mask by creating the channel dimension and repeating the mask"
    return np.repeat(mask[..., np.newaxis], num_channels, axis=-1)

def get_almost_white_pixel_mask(arr_img, threshold):
    "arr_img: (..., C) with values in [0, 1]"
    return np.all(arr_img > threshold, axis=-1)

def get_almost_white_pixels(arr_img, threshold=0.98):
    "arr_img: (..., C) with values in [0, 1]"
    num_channels = arr_img.shape[-1]
    mask = expand_mask(get_almost_white_pixel_mask(arr_img, threshold), num_channels)
    return arr_img[mask].reshape(-1, num_channels)

def get_fg_bg_pixels(arr_img):
    "arr_img: (..., C) with values in [0, 1]"
    num_channels = arr_img.shape[-1]
    bg_mask = expand_mask(get_white_pixels_mask(arr_img), num_channels)
    fg_mask = np.logical_not(bg_mask)
    return arr_img[fg_mask].reshape(-1, num_channels), arr_img[bg_mask].reshape(-1, num_channels)

def get_num_pixels(arr_img):
    "arr_img: (..., C) with values in [0, 1]"
    if arr_img.ndim <= 1:
        return 0
    return np.prod(arr_img.shape[:-1])

def is_scene_valid(
        scene_imgs, 
        bg_fg_threshold=0.1, #disable if 0
        almost_white_pixel_value_threshold=0.7,
        almost_white_pixel_ratio_threshold=0.9 # disable if 1
    ):
    "scene_imgs: (..., C) with values in [0, 1]"

    fg_pixels, bg_pixels = get_fg_bg_pixels(scene_imgs)


    # Filter out fully white scenes
    num_fg_pixels = get_num_pixels(fg_pixels)
    num_bg_pixels = get_num_pixels(bg_pixels)
    num_pixels = num_fg_pixels + num_bg_pixels

    validity = True
    why = []

    if num_fg_pixels == 0:
        validity = False
        why.append("No object")
    
    # Filter out scenes with too many bg pixels (really small objects)
    if num_fg_pixels < bg_fg_threshold * num_pixels:
        validity = False
        why.append("Object too small")
    
    almost_white_pixels = get_almost_white_pixels(fg_pixels, threshold=almost_white_pixel_value_threshold)
    num_almost_white_pixels = get_num_pixels(almost_white_pixels)
    if num_almost_white_pixels > almost_white_pixel_ratio_threshold * num_fg_pixels:
        validity = False
        why.append("Too many almost white pixels")
    
    return validity, why
    
    
if __name__ == "__main__":
    # --- args ---- 
    parser = argparse.ArgumentParser(description='Copy valid scenes of an objaverse dataset')
    parser.add_argument('--source', type=str, default="objaverse/medium", help='source dataset')
    parser.add_argument('--target', type=str, default="objaverse/medium_curated", help='target dataset')
    args = parser.parse_args()

    subset_to_copy_from = 'raw'
    subset_to_read_from = 'square128'
    # --- end of args ---- 


    DATA_DIR = Path("path/to/data/dir")


    n_scenes = np.inf
    dataset_config = Prodict.from_dict({
            'name': args.source,
            "subset": subset_to_read_from,
            "scene_repartition": { 
                'objaverse': (0, n_scenes)
            },
            "few_view_factor": 1.0,
        })

    mv_dset = DataReader(dataset_config)

    c = 0 
    os.system(f"rm -rf {DATA_DIR / args.target}")
    for scene_idx, scene_path in tqdm.tqdm(enumerate(mv_dset.path_of_scenes)):
        views = sample_views(scene_idx, n_views=10)
        validity, why = is_scene_valid(
            views,
            bg_fg_threshold=0.05, #disable if 0
            almost_white_pixel_ratio_threshold=1 # disable if 1
        )

        if validity:
            scene_id = scene_path.name
            copy_from = DATA_DIR / args.source / subset_to_copy_from / scene_id
            copy_to = DATA_DIR / args.target / subset_to_copy_from 
            os.makedirs(copy_to, exist_ok=True)
            os.system(f"cp -r {copy_from} {copy_to}")
            c += 1


    print(f"copied {c} scenes out of {len(mv_dset.path_of_scenes)}")
