import os
import PIL.Image
from prodict import Prodict
import numpy as np
import matplotlib.pyplot as plt
from pathlib import Path
import json
import random
import torchvision.transforms as tr
import torch
import tqdm
import argparse
import sys

if __name__ == '__main__':
    sys.path.append(str(Path(__file__).resolve().parents[3]))

from datasets.dataset import DataReader


def normalize(img_arr):
    return img_arr * 0.5 + 0.5

def sample_views(idx_obj, n_views=10):
    "Returns a array of views of a scene specified by idx_obj. Retunred array has shape (n_views, H, W, C) with values in [0, 1]"
    views = []
    for i in range(n_views):
        img_arr = mv_dset.get_image(idx_obj, i)['img'].numpy()
        views.append(normalize(img_arr).transpose(1, 2, 0))
    return np.stack(views, axis=0)

def get_white_pixels_mask(arr_img):
    """args:
        arr_img: (..., C) with values in [0, 1]
    returns: 
        mask (...) with values in {0, 1}. 1 if pixel is white, 0 otherwise
    """
    return np.all(arr_img == 1, axis=-1)

def expand_mask(mask, num_channels=3):
    "expands a mask by creating the channel dimension and repeating the mask"
    return np.repeat(mask[..., np.newaxis], num_channels, axis=-1)

def get_almost_white_pixel_mask(arr_img, threshold):
    "arr_img: (..., C) with values in [0, 1]"
    return np.all(arr_img > threshold, axis=-1)

def get_almost_white_pixels(arr_img, threshold=0.98):
    "arr_img: (..., C) with values in [0, 1]"
    num_channels = arr_img.shape[-1]
    mask = expand_mask(get_almost_white_pixel_mask(arr_img, threshold), num_channels)
    return arr_img[mask].reshape(-1, num_channels)

def get_fg_bg_pixels(arr_img):
    "arr_img: (..., C) with values in [0, 1]"
    num_channels = arr_img.shape[-1]
    bg_mask = expand_mask(get_white_pixels_mask(arr_img), num_channels)
    fg_mask = np.logical_not(bg_mask)
    return arr_img[fg_mask].reshape(-1, num_channels), arr_img[bg_mask].reshape(-1, num_channels)

def get_num_pixels(arr_img):
    "arr_img: (..., C) with values in [0, 1]"
    if arr_img.ndim <= 1:
        return 0
    return np.prod(arr_img.shape[:-1])

def is_scene_valid(
        scene_imgs, 
        bg_fg_threshold=0.1, #disable if 0
        almost_white_pixel_value_threshold=0.7,
        almost_white_pixel_ratio_threshold=0.9 # disable if 1
    ):
    "scene_imgs: (..., C) with values in [0, 1]"

    fg_pixels, bg_pixels = get_fg_bg_pixels(scene_imgs)


    # Filter out fully white scenes
    num_fg_pixels = get_num_pixels(fg_pixels)
    num_bg_pixels = get_num_pixels(bg_pixels)
    num_pixels = num_fg_pixels + num_bg_pixels

    validity = True
    why = []

    if num_fg_pixels == 0:
        validity = False
        why.append("No object")
    
    # Filter out scenes with too many bg pixels (really small objects)
    if num_fg_pixels < bg_fg_threshold * num_pixels:
        validity = False
        why.append("Object too small")
    
    almost_white_pixels = get_almost_white_pixels(fg_pixels, threshold=almost_white_pixel_value_threshold)
    num_almost_white_pixels = get_num_pixels(almost_white_pixels)
    if num_almost_white_pixels > almost_white_pixel_ratio_threshold * num_fg_pixels:
        validity = False
        why.append("Too many almost white pixels")
    
    return validity, why


class SceneSelector:

    def __init__(self):
        self.valid_scenes = []
        self.invalid_scenes = []
        self.current_scene_idx = 0

    def next(self) : 
        self.current_scene_idx += 1

    def display_scene(self, idx_obj, n_views=5):
        "Displays the scene specified by idx_obj"
        views = sample_views(idx_obj, n_views)
        fig, axes = plt.subplots(1, n_views, figsize=(n_views * 3, 3))
        for i, ax in enumerate(axes):
            ax.imshow(views[i])
            ax.axis('off')
        plt.show()

    def main(self):
        print("""Scenes will be displayed. For each scene, press space if its valid, otherwise press D for delete.""")
        print("Press Q to quit.")
        while True:
            validity, why = is_scene_valid(
                sample_views(self.current_scene_idx),
                bg_fg_threshold=0.1, 
                almost_white_pixel_ratio_threshold=0.8
            )
            if validity:
                self.display_scene(self.current_scene_idx) 
                user_input = input()
                if user_input == " ":
                    self.valid_scenes.append(self.current_scene_idx)
                    print(f"Number of valid scenes: {len(self.valid_scenes)}")
                    self.next()
                elif user_input.lower() == "d":
                    self.invalid_scenes.append(self.current_scene_idx)
                    self.next()
                elif user_input.lower() == "q":
                    break
                else:
                    print("Invalid input. Please press space or D.")
            else:
                self.invalid_scenes.append(self.current_scene_idx)
                self.next()
        
if __name__ == "__main__":
    # --- args ---- 
    source="objaverse/medium"
    target="objaverse/medium_manually_curated"

    subset_to_copy_from = 'raw'
    subset_to_read_from = 'square128'
    # --- end of args ---- 


    DATA_DIR = Path("path/to/data/dir")


    n_scenes = np.inf
    dataset_config = Prodict.from_dict({
            'name': source,
            "subset": subset_to_read_from,
            "scene_repartition": { 
                'objaverse': (0, n_scenes)
            },
            "few_view_factor": 1.0,
        })

    #mv_dset = DataReader(dataset_config)

    scene_selector = SceneSelector()


    scene_selector.valid_scenes = [0,
 1,
 3,
 5,
 7,
 8,
 9,
 14,
 29,
 34,
 35,
 41,
 48,
 50,
 52,
 55,
 57,
 65,
 79,
 87,
 88,
 89,
 90,
 91,
 94,
 96,
 97,
 98,
 100,
 102,
 109,
 111,
 128,
 129,
 131,
 133,
 134,
 136,
 137,
 139,
 141,
 142,
 143,
 144,
 145,
 146,
 147,
 148,
 149,
 150,
 152,
 153,
 155,
 157,
 158,
 159,
 166,
 173,
 175,
 177,
 178,
 179,
 180,
 182,
 183,
 184,
 185,
 187,
 188,
 196,
 197,
 198,
 199,
 200,
 201,
 202,
 208,
 209,
 211,
 217,
 219,
 223,
 226,
 232,
 233,
 240,
 241,
 252,
 254,
 256,
 260,
 262,
 266,
 268,
 273,
 277,
 278,
 280,
 286,
 289,
 291,
 292,
 296,
 305,
 309,
 326,
 329,
 330,
 331,
 335,
 343,
 349,
 353,
 360,
 361,
 363,
 371,
 377,
 380,
 383,
 384,
 392,
 393,
 394,
 400,
 402,
 403,
 404,
 410,
 411,
 420,
 422,
 427,
 430,
 435,
 442,
 450,
 452,
 454,
 456,
 458,
 461,
 462,
 465,
 467,
 468,
 469,
 470,
 471,
 480,
 486,
 488,
 490,
 492,
 496,
 498,
 503,
 504,
 508,
 510,
 519,
 521,
 522,
 524,
 525,
 535,
 539,
 542,
 543,
 544,
 551,
 554,
 560,
 561,
 563,
 575,
 576,
 581,
 585,
 586,
 587,
 589,
 590,
 595,
 596,
 597,
 598,
 600,
 601,
 603,
 606,
 607,
 608,
 609,
 610,
 611,
 615,
 621,
 624,
 627,
 630,
 631,
 632,
 633,
 636,
 637,
 638,
 639,
 641,
 642,
 645,
 650,
 651,
 653,
 654,
 661,
 666,
 668,
 676,
 678,
 679,
 681,
 683,
 685,
 697,
 698,
 707,
 716,
 722,
 723,
 732,
 739,
 740,
 750,
 752,
 753,
 754,
 755,
 760,
 763,
 765,
 766,
 768,
 773,
 781,
 783,
 787,
 791,
 795,
 799,
 800,
 801,
 802,
 805,
 809,
 810,
 812,
 814,
 821,
 822,
 825,
 829,
 832,
 838,
 839,
 843,
 844,
 845,
 846,
 848,
 850,
 851,
 854,
 856,
 869,
 877,
 879,
 881,
 882,
 885,
 886,
 887,
 888,
 890,
 892,
 893,
 900,
 902,
 903,
 906,
 907,
 909,
 910,
 912,
 913,
 914,
 916,
 917,
 920,
 921,
 922,
 926,
 934,
 940,
 942,
 944,
 947,
 950,
 953,
 954,
 960,
 964,
 965,
 966,
 967,
 970,
 982,
 985,
 986,
 988,
 989,
 990,
 991,
 997,
 1006,
 1009,
 1010,
 1015,
 1045,
 1047,
 1049,
 1050,
 1052,
 1055,
 1059,
 1070,
 1076,
 1081,
 1084,
 1085,
 1086,
 1087,
 1090,
 1092,
 1107,
 1109,
 1110,
 1111,
 1130,
 1131,
 1132,
 1134,
 1140,
 1142,
 1143,
 1145,
 1152,
 1155,
 1158,
 1160,
 1162,
 1163,
 1183,
 1184,
 1185,
 1186,
 1189,
 1197,
 1198,
 1202,
 1204,
 1208,
 1211,
 1214,
 1217,
 1219,
 1224,
 1225,
 1231,
 1232,
 1237,
 1241,
 1244,
 1246,
 1247,
 1248,
 1255,
 1258,
 1269,
 1271,
 1275,
 1290,
 1294,
 1298,
 1300,
 1302,
 1304,
 1305,
 1307,
 1308,
 1313,
 1319,
 1321,
 1326,
 1328,
 1331,
 1343,
 1344,
 1346,
 1349,
 1357,
 1358,
 1359,
 1362,
 1365,
 1369,
 1372,
 1374,
 1375,
 1379,
 1381,
 1385,
 1386,
 1390,
 1391,
 1404,
 1406,
 1410,
 1412,
 1420,
 1434,
 1435,
 1436,
 1441,
 1444,
 1456,
 1462,
 1464,
 1465,
 1471,
 1473,
 1474,
 1479,
 1481,
 1482,
 1484,
 1485,
 1490,
 1491,
 1497,
 1500,
 1501,
 1502,
 1504,
 1507,
 1511,
 1515,
 1519,
 1521,
 1522,
 1526,
 1529,
 1534,
 1537,
 1546,
 1552,
 1553,
 1560,
 1564,
 1565,
 1566,
 1567,
 1570,
 1574,
 1577,
 1582,
 1588,
 1597,
 1598,
 1600,
 1609,
 1614,
 1616,
 1617,
 1619,
 1621,
 1622,
 1634,
 1657,
 1658,
 1662,
 1665,
 1672,
 1680,
 1682,
 1692,
 1694,
 1699,
 1707,
 1711,
 1721,
 1722,
 1723,
 1728,
 1730,
 1739,
 1748,
 1749,
 1750,
 1754,
 1760,
 1768,
 1777,
 1778,
 1782,
 1786,
 1794,
 1798,
 1800,
 1813,
 1815,
 1817,
 1838,
 1843,
 1844,
 1847]
    scene_selector.current_scene_idx = 1848
    #scene_selector.main()



    # save chose scenes in a txt file
    with open("chosen_scenes.txt", 'w') as f:
        for scene in scene_selector.valid_scenes:
            f.write(f"{scene}\n")
