# Copyright 2022 the Regents of the University of California, Nerfstudio Team and contributors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Put all the method implementations in one location.
"""

from __future__ import annotations

from collections import OrderedDict
from typing import Dict, Union

import tyro

from nerfstudio.cameras.camera_optimizers import CameraOptimizerConfig
from nerfstudio.configs.external_methods import ExternalMethodDummyTrainerConfig, get_external_methods
from nerfstudio.data.dataparsers.nerfstudio_dataparser import NerfstudioDataParserConfig
from nerfstudio.data.dataparsers.blender_dataparser import BlenderDataParserConfig
from nerfstudio.engine.optimizers import AdamOptimizerConfig, RAdamOptimizerConfig
from nerfstudio.engine.schedulers import (
    CosineDecaySchedulerConfig,
    ExponentialDecaySchedulerConfig,
    MultiStepSchedulerConfig,
)
from nerfstudio.engine.trainer import TrainerConfig
from nerfstudio.field_components.temporal_distortions import TemporalDistortionKind
from nerfstudio.fields.sdf_field import SDFFieldConfig
from nerfstudio.models.depth_nerfacto import DepthNerfactoModelConfig
from nerfstudio.models.generfacto import GenerfactoModelConfig
from nerfstudio.models.instant_ngp import InstantNGPModelConfig
from nerfstudio.models.mipnerf import MipNerfModel
from nerfstudio.models.nerfacto import NerfactoModelConfig
from nerfstudio.models.neus import NeuSModelConfig
from nerfstudio.models.neus_facto import NeuSFactoModelConfig
from nerfstudio.models.semantic_nerfw import SemanticNerfWModelConfig
from nerfstudio.models.splatfacto import SplatfactoModelConfig
from nerfstudio.models.tensorf import TensoRFModelConfig
from nerfstudio.models.vanilla_nerf import NeRFModel, VanillaModelConfig
from nerfstudio.pipelines.base_pipeline import VanillaPipelineConfig
from nerfstudio.pipelines.dynamic_batch import DynamicBatchPipelineConfig
from nerfstudio.plugins.registry import discover_methods
from nerfstudio.plugins.types import MethodSpecification
from kplanes_nerfstudio.kplanes.kplanes_configs import  kplanes_method
from kplanes_nerfstudio.kplanes.kplanes import KPlanesModelConfig

# our imports
from custom_code.custom_pipeline import CustomPipelineConfig
from custom_code.method import MethodConfig
from custom_code.custom_datamanager import CustomFullImageDatamanagerConfig

def get_dataset_specific_args(YAML_CONFIG):

    if YAML_CONFIG.dataset.subset.startswith("blender"):
        dataset_type = 'blender'
    elif YAML_CONFIG.dataset.name.startswith("co3d"):
        dataset_type = 'co3d'
    elif YAML_CONFIG.dataset.name.startswith("shapenet"):
        dataset_type = 'shapenet'
    elif YAML_CONFIG.dataset.name.startswith("objaverse"):
        dataset_type = 'objaverse'
    elif YAML_CONFIG.dataset.name.startswith("basel_faces"):
        dataset_type = 'basel_faces'
    else:
        raise ValueError(f"Unknown dataset type could not be determined")
    

    if dataset_type == 'blender':
        scene_bbox_scale_factor = 0.32
        dataparser_config = BlenderDataParserConfig(
            scale_factor = scene_bbox_scale_factor
        )
        near_plane = 0.2
        far_plane = 2.4
    elif dataset_type == 'co3d':
        dataparser_config = NerfstudioDataParserConfig(
                    train_split_fraction=0.9,
                    downscale_factor=YAML_CONFIG.dataset.downscale_factor
                )
        near_plane = 0.2
        far_plane = 4.0
    elif dataset_type == 'shapenet':
        dataparser_config = NerfstudioDataParserConfig(
                    train_split_fraction=0.9,
                    downscale_factor=YAML_CONFIG.dataset.downscale_factor
                )
        near_plane = 0.2
        far_plane = 2.4
    elif dataset_type == 'basel_faces':
        dataparser_config = NerfstudioDataParserConfig(
                train_split_fraction=0.9,
                downscale_factor=YAML_CONFIG.dataset.downscale_factor
            )
        near_plane = 0.2
        far_plane = 2.4

        dataparser_config = NerfstudioDataParserConfig(
                    train_split_fraction=0.9,
                    downscale_factor=YAML_CONFIG.dataset.downscale_factor
                )
        near_plane = 0.2
        far_plane = 2.4
    elif dataset_type == 'objaverse':
        dataparser_config = NerfstudioDataParserConfig(
                    train_split_fraction=0.9,
                    downscale_factor=YAML_CONFIG.dataset.downscale_factor
                )
        near_plane = 0.2
        far_plane = 2.4
    else:
        raise ValueError(f"Unknown dataset type: {dataset_type}")
    
    return dataparser_config, near_plane, far_plane

def get_custom_method_configs(YAML_CONFIG): 
    "Configuration for the different methods in a NeRFStudio style, according to the YAML_CONFIG."

    lr_factor = YAML_CONFIG.optim.nerf.lr_factor

    # dataset specific args
    dataparser_config, near_plane, far_plane = get_dataset_specific_args(YAML_CONFIG)
    
    # we use the same datamanager for all methods
    datamanager_config = CustomFullImageDatamanagerConfig(
                dataparser=dataparser_config,
                cache_latents=YAML_CONFIG.dataset.cache_latents
            )
    custom_method_configs: Dict[str, Union[TrainerConfig, ExternalMethodDummyTrainerConfig]] = {}

    # nerfacto
    method_name = "nerfacto"
    custom_method_configs[method_name] = MethodConfig(
        method_name=method_name,
        mixed_precision=True, 
        pipeline=CustomPipelineConfig(
            datamanager=datamanager_config,
            model=NerfactoModelConfig(
                eval_num_rays_per_chunk=1 << 15,
                average_init_density=0.01,
                camera_optimizer=CameraOptimizerConfig(mode="off"),
                near_plane=near_plane,
                far_plane=far_plane,
                use_appearance_embedding=YAML_CONFIG.custom_settings.nerfacto.use_appearance_embedding, # HARDCODED: Custom modif, otherwise blue renderings
                use_average_appearance_embedding=YAML_CONFIG.custom_settings.nerfacto.use_average_appearance_embedding,
            ),
        ),
        optimizers={
            "proposal_networks": {
                "optimizer": AdamOptimizerConfig(lr=1e-2*lr_factor[method_name], eps=1e-15),
                "scheduler": ExponentialDecaySchedulerConfig(lr_final=0.0001*lr_factor[method_name], max_steps=200000),
            },
            "fields": {
                "optimizer": AdamOptimizerConfig(lr=1e-2*lr_factor[method_name], eps=1e-15),
                "scheduler": ExponentialDecaySchedulerConfig(lr_final=0.0001*lr_factor[method_name], max_steps=200000),
            },
            "camera_opt": {
                "optimizer": AdamOptimizerConfig(lr=1e-3*lr_factor[method_name], eps=1e-15),
                "scheduler": ExponentialDecaySchedulerConfig(lr_final=1e-4*lr_factor[method_name], max_steps=5000),
            },
        }
    )

    # vanilla nerf
    method_name = "vanilla_nerf"
    custom_method_configs[method_name] = MethodConfig(
        method_name=method_name,
        pipeline=CustomPipelineConfig(
            datamanager=datamanager_config,
            model=VanillaModelConfig(
                _target=NeRFModel,
            ),
        ),
        optimizers={
            "fields": {
                "optimizer": RAdamOptimizerConfig(lr=5e-4*lr_factor[method_name], eps=1e-08),
                "scheduler": None,
            },
            "temporal_distortion": {
                "optimizer": RAdamOptimizerConfig(lr=5e-4*lr_factor[method_name], eps=1e-08),
                "scheduler": None,
            },
        },
    )

    method_name = "tensorf"
    custom_method_configs[method_name] = MethodConfig(
        method_name=method_name,
        mixed_precision=False,
        pipeline=CustomPipelineConfig(
            datamanager=datamanager_config,
            model=TensoRFModelConfig(
                regularization="tv",
                camera_optimizer=CameraOptimizerConfig(mode="off"),
            ),
        ),
        optimizers={
            "fields": {
                "optimizer": AdamOptimizerConfig(lr=0.001*lr_factor[method_name]),
                "scheduler": ExponentialDecaySchedulerConfig(lr_final=0.0001*lr_factor[method_name], max_steps=30000),
            },
            "encodings": {
                "optimizer": AdamOptimizerConfig(lr=0.02*lr_factor[method_name]),
                "scheduler": ExponentialDecaySchedulerConfig(lr_final=0.002*lr_factor[method_name], max_steps=30000),
            },
            "camera_opt": {
                "optimizer": AdamOptimizerConfig(lr=1e-4*lr_factor[method_name], eps=1e-15),
                "scheduler": ExponentialDecaySchedulerConfig(lr_final=1e-5, max_steps=5000),
            },
        },
    )

    method_name = "kplanes"
    custom_method_configs[method_name] = MethodConfig(
        method_name=method_name,
        mixed_precision=True,
        pipeline=CustomPipelineConfig(
            datamanager=datamanager_config,
            model=KPlanesModelConfig(
                eval_num_rays_per_chunk=1 << 15,
                grid_base_resolution=[128, 128, 128],
                grid_feature_dim=32,
                multiscale_res=[1, 2, 4],
                proposal_net_args_list=[
                    {"num_output_coords": 8, "resolution": [128, 128, 128]},
                    {"num_output_coords": 8, "resolution": [256, 256, 256]}
                ],
                loss_coefficients={
                    "interlevel": 1.0,
                    "distortion": 0.01,
                    "plane_tv": 0.01,
                    "plane_tv_proposal_net": 0.0001,
                },
                near_plane = near_plane, # HARDCODED
                far_plane = far_plane, # HARDCODED
            ),
        ),
        optimizers={
            "proposal_networks": {
                "optimizer": AdamOptimizerConfig(lr=1e-2*lr_factor[method_name], eps=1e-12),
                "scheduler": CosineDecaySchedulerConfig(warm_up_end=512, max_steps=30000),
            },
            "fields": {
                "optimizer": AdamOptimizerConfig(lr=1e-2*lr_factor[method_name], eps=1e-12),
                "scheduler": CosineDecaySchedulerConfig(warm_up_end=512, max_steps=30000),
            },
        },
    )

    method_name = "instant_ngp"
    custom_method_configs[method_name] = MethodConfig(
        method_name=method_name,
        mixed_precision=True,
        pipeline=CustomPipelineConfig(
            datamanager=datamanager_config,
            model=InstantNGPModelConfig(
                eval_num_rays_per_chunk=8192,
                use_appearance_embedding=YAML_CONFIG.custom_settings.instant_ngp.use_appearance_embedding,
            ),
        ),
        optimizers={
            "fields": {
                "optimizer": AdamOptimizerConfig(lr=1e-2*lr_factor[method_name], eps=1e-15),
                "scheduler": ExponentialDecaySchedulerConfig(lr_final=0.0001*lr_factor[method_name], max_steps=200000),
            }
        },
    )

    method_name = "instant_ngp_bounded"
    custom_method_configs[method_name] = MethodConfig(
        method_name=method_name,
        mixed_precision=True,
        pipeline=CustomPipelineConfig(
            datamanager=datamanager_config,
            model=InstantNGPModelConfig(
                eval_num_rays_per_chunk=8192,
                grid_levels=1,
                alpha_thre=0.0,
                cone_angle=0.0,
                disable_scene_contraction=True,
                near_plane=0.01,
            ),
        ),
        optimizers={
            "fields": {
                "optimizer": AdamOptimizerConfig(lr=1e-2*lr_factor[method_name], eps=1e-15),
                "scheduler": ExponentialDecaySchedulerConfig(lr_final=0.0001*lr_factor[method_name], max_steps=200000),
            }
        },
    )

    method_name = "splatfacto"
    custom_method_configs[method_name] = TrainerConfig(
        method_name=method_name,
        steps_per_eval_image=100,
        steps_per_eval_batch=0,
        steps_per_save=2000,
        steps_per_eval_all_images=1000,
        max_num_iterations=30000,
        mixed_precision=False,
        pipeline=VanillaPipelineConfig(
            datamanager=datamanager_config, #TODO: check load_3D_points
            
            # FullImageDatamanagerConfig(
            #     dataparser=NerfstudioDataParserConfig(load_3D_points=True),
            #     cache_images_type="uint8",
            # ),

            model=SplatfactoModelConfig(),
        ),
        optimizers={
            "means": {
                "optimizer": AdamOptimizerConfig(lr=1.6e-4*lr_factor[method_name], eps=1e-15),
                "scheduler": ExponentialDecaySchedulerConfig(
                    lr_final=1.6e-6,
                    max_steps=30000,
                ),
            },
            "features_dc": {
                "optimizer": AdamOptimizerConfig(lr=0.0025*lr_factor[method_name], eps=1e-15),
                "scheduler": None,
            },
            "features_rest": {
                "optimizer": AdamOptimizerConfig(lr=0.0025*lr_factor[method_name] / 20, eps=1e-15),
                "scheduler": None,
            },
            "opacities": {
                "optimizer": AdamOptimizerConfig(lr=0.05*lr_factor[method_name], eps=1e-15),
                "scheduler": None,
            },
            "scales": {
                "optimizer": AdamOptimizerConfig(lr=0.005*lr_factor[method_name], eps=1e-15),
                "scheduler": None,
            },
            "quats": {"optimizer": AdamOptimizerConfig(lr=0.001*lr_factor[method_name], eps=1e-15), "scheduler": None},
            "camera_opt": {
                "optimizer": AdamOptimizerConfig(lr=1e-4*lr_factor[method_name], eps=1e-15),
                "scheduler": ExponentialDecaySchedulerConfig(
                    lr_final=5e-7*lr_factor[method_name], max_steps=30000, warmup_steps=1000, lr_pre_warmup=0
                ),
            },
        },
    )   

    return custom_method_configs






