from __future__ import annotations

import dataclasses
import functools
import os
import time
from diffusers import AutoencoderKL
import yaml
from datetime import datetime
from collections import defaultdict
from dataclasses import dataclass, field
from pathlib import Path
from threading import Lock
from typing import Any, DefaultDict, Dict, List, Literal, Optional, Tuple, Type, cast
import torch
from rich import box, style
from rich.panel import Panel
from rich.table import Table
from torch.cuda.amp.grad_scaler import GradScaler
from prodict import Prodict
import wandb
from tqdm import trange

from nerfstudio.configs.base_config import InstantiateConfig, LoggingConfig, MachineConfig, ViewerConfig
from nerfstudio.engine.callbacks import TrainingCallback, TrainingCallbackAttributes, TrainingCallbackLocation
from nerfstudio.engine.optimizers import OptimizerConfig, Optimizers
from nerfstudio.engine.schedulers import SchedulerConfig, Scheduler
from nerfstudio.utils import profiler
from nerfstudio.utils.decorators import check_eval_enabled, check_main_thread, check_viewer_enabled
from nerfstudio.utils.misc import step_check
from nerfstudio.utils.rich_utils import CONSOLE
from nerfstudio.configs.config_utils import to_immutable_dict
# our imports
from custom_code.custom_pipeline import CustomPipeline, CustomPipelineConfig
from custom_code.utils import set_requires_grad, step_in_range, TanhNormalizer


def transform_loss_dict(loss_dict, mode='latent'):
    if mode != 'latent':
        return loss_dict
    logged_dict = {}
    for key, value in loss_dict.items():
        if 'rgb' in key:
            logged_dict[key.replace('rgb', 'latent')] = value
        else:
            logged_dict[key] = value
    return logged_dict

@dataclass
class MethodConfig(InstantiateConfig):
    """Configuration for training regimen"""

    _target: Type = field(default_factory=lambda: Method)
    """target class to instantiate"""
    
    method_name: Optional[str] = "method_name"
    method_idx: int = -1
    scene_idx: int = -1

    """Method name. Required to set in python or via cli"""
    optimizers: Dict[str, Any] = to_immutable_dict(
        {
            "fields": {
                "optimizer": OptimizerConfig(),
                "scheduler": SchedulerConfig(),
            }
        }
    )
    """Dictionary of optimizer groups and their schedulers"""
    pipeline: CustomPipelineConfig = field(default_factory=CustomPipeline)
    """Pipeline configuration"""
    mixed_precision: bool = False
    """Whether or not to use mixed precision for training."""
    use_grad_scaler: bool = False
    """Use gradient scaler even if the automatic mixed precision is disabled."""
    gradient_accumulation_steps: Dict[str, int] = field(default_factory=lambda: {})
    """Number of steps to accumulate gradients over. Contains a mapping of {param_group:num}"""
    

class Method:

    pipeline: CustomPipeline
    optimizers: Optimizers
    callbacks: List[TrainingCallback]

    def __init__(
            self, 
            config: MethodConfig, 
            vae: AutoencoderKL,
            normalizer: TanhNormalizer,
            img_size: int,
            render_img_size: int,
            eval_funcs: Dict[str, torch.nn.Module],
            device: str,
            local_rank: int = 0,
            test_mode: Literal["test", "val", "inference"] = "val"
        ):

        self.train_lock = Lock()
        self.config = config
        self.local_rank = local_rank
        self.world_size = 1
        self.device = device
        self.method_name = self.config.method_name
        self.scene_idx = self.config.scene_idx
        self.method_idx = self.config.method_idx

        if self.device == "cuda":
            self.device += f":{local_rank}"
        self.mixed_precision: bool = self.config.mixed_precision
        self.use_grad_scaler: bool = self.mixed_precision or self.config.use_grad_scaler
        self.gradient_accumulation_steps: DefaultDict = defaultdict(lambda: 1)
        self.gradient_accumulation_steps.update(self.config.gradient_accumulation_steps)
        
        self.grad_scaler = GradScaler(enabled=self.use_grad_scaler)
        self.pipeline = self.config.pipeline.setup(
            vae=vae,
            normalizer=normalizer,
            img_size=img_size,
            render_img_size=render_img_size,
            eval_funcs=eval_funcs,
            device=self.device,
            test_mode=test_mode,
            local_rank=self.local_rank,
            grad_scaler=self.grad_scaler,
        )

        # optimizers
        self.setup_nerf_optim()
        self.disable_nerf_training()
        self.callbacks = self.pipeline.get_training_callbacks(
            TrainingCallbackAttributes(
                optimizers=self.optimizers, grad_scaler=self.grad_scaler, pipeline=self.pipeline, trainer=self
            )
        )

    @property
    def method_id(self):
        return f"{self.method_name}_{self.method_idx}"
        
    def setup_nerf_optim(self, print_n_params=False):
        optimizer_config = self.config.optimizers.copy()
        param_groups = self.pipeline.get_param_groups()
        self.optimizers = Optimizers(optimizer_config, param_groups)
        self.optimizers_initial_state = {
            'optimizers' : {
                key: optimizer.state_dict() for key, optimizer in self.optimizers.optimizers.items()
            },
            'schedulers': {
                key: scheduler.state_dict() for key, scheduler in self.optimizers.schedulers.items()
            }
        }

        if print_n_params:
            self.n_nerf_param = 0
            for param_group in param_groups.values():
                for param in param_group:
                    self.n_nerf_param += param.numel()
            print(f"METHOD_NAME: {self.config.method_name} - Number of nerf parameters: {self.n_nerf_param}")

    def reset_nerf_optimizers_states(self):
        for key, optimizer in self.optimizers.optimizers.items():
            optimizer.load_state_dict(self.optimizers_initial_state['optimizers'][key])
        for key, scheduler in self.optimizers.schedulers.items():
            scheduler.load_state_dict(self.optimizers_initial_state['schedulers'][key])

    def enable_nerf_training(self) -> None:
        self.train_nerf = True        
        param_groups = self.pipeline.get_param_groups()
        for param_group in param_groups.values():
            for param in param_group:
                param.requires_grad_(True)

    def disable_nerf_training(self) -> None: 
        self.train_nerf = False
        param_groups = self.pipeline.get_param_groups()
        for param_group in param_groups.values():
            for param in param_group:
                param.requires_grad_(False)

    def training_iteration(self, step):
        self.pipeline.train()
        self.callback_before(step)
        loss, loss_dict, metrics_dict = self.forward_pass(step)
        self.grad_scaler.scale(loss).backward()  
        self.optimizer_scheduler_step(step)
        self.callback_after(step)
        return loss, loss_dict, metrics_dict

    def callback_before(self, step):
        for callback in self.callbacks:
            callback.run_callback_at_location(
                step, location=TrainingCallbackLocation.BEFORE_TRAIN_ITERATION
            )

    def callback_after(self, step):
        for callback in self.callbacks:
            callback.run_callback_at_location(
                step, location=TrainingCallbackLocation.AFTER_TRAIN_ITERATION
            )

    def forward_pass(self, step):

        # forward pass
        cpu_or_cuda_str: str = self.device.split(":")[0]
        cpu_or_cuda_str = "cpu" if cpu_or_cuda_str == "mps" else cpu_or_cuda_str

        if self.train_nerf:
            needs_zero = [
                group for group in self.optimizers.parameters.keys() if step % self.gradient_accumulation_steps[group] == 0
            ]
            self.optimizers.zero_grad_some(needs_zero)

        # ---- backward here ----
        with torch.autocast(device_type=cpu_or_cuda_str, enabled=self.mixed_precision):
            _, loss_dict, metrics_dict = self.pipeline.get_train_loss_dict(step=step)
            assert len(loss_dict) != 0, "You should optimize at least one loss"
            loss = functools.reduce(torch.add, loss_dict.values())

        return loss, loss_dict, metrics_dict
    
    def optimizer_scheduler_step(self, step):

        if self.train_nerf:
            needs_step = [
                group
                for group in self.optimizers.parameters.keys()
                if step % self.gradient_accumulation_steps[group] == self.gradient_accumulation_steps[group] - 1
            ]
            self.optimizers.optimizer_scaler_step_some(self.grad_scaler, needs_step)
            scale = self.grad_scaler.get_scale()
            self.grad_scaler.update()
            # If the gradient scaler is decreased, no optimization step is performed so we should not step the scheduler.
            if scale <= self.grad_scaler.get_scale():
                self.optimizers.scheduler_step_all(step)

        # if step_check(step, self.config.logging.steps_per_log, run_at_zero=True):
        #     to_log = {
        #         'training_step': step,
        #         'train': {
        #             'total_loss': loss,
        #             'losses': transform_loss_dict(loss_dict, mode=TRAINING_MODE),
        #         },
        #         "status" : {
        #                 "train_nerf": int(self.train_nerf),
        #                 "train_encoder": int(self.train_encoder),
        #                 "train_decoder": int(self.train_decoder),
        #                 "loss_e": int(step_in_range(step,  self.config.yaml_config.optim.loss_e_range)),
        #                 "loss_d": int(step_in_range(step,  self.config.yaml_config.optim.loss_d_range)),
        #                 "loss_ae": int(step_in_range(step,  self.config.yaml_config.optim.loss_ae_range)),
        #                 'GPU memory': torch.cuda.max_memory_allocated() / (1024**2)
        #         },
        #     }
        #     if 'metrics_e' in metrics_dict.keys(): 
        #         to_log['train']['metrics_e'] = metrics_dict['metrics_e']
        #     if 'metrics_ae' in metrics_dict.keys(): 
        #         to_log['train']['metrics_ae'] = metrics_dict['metrics_ae']
        #     if 'metrics_d' in metrics_dict.keys():
        #         to_log['train']['metrics_d'] = metrics_dict['metrics_d']
        #     wandb.log(to_log)

        # # Do not perform evaluation if there are no validation images
        # if self.pipeline.datamanager.eval_dataset:
        #     self.eval_iteration(step)

        # if step_check(step, self.config.steps_per_save):
        #     self.save_checkpoint(step)

    def load_state_dict(self, loaded_state, step):
            self.pipeline.load_pipeline(loaded_state["pipeline"], step)
            # self.optimizers.load_optimizers(loaded_state["optimizers"])
            # if "schedulers" in loaded_state and self.config.load_scheduler:
            #     self.optimizers.load_schedulers(loaded_state["schedulers"])
            # self.grad_scaler.load_state_dict(loaded_state["scalers"])
