from __future__ import annotations

import math
import dataclasses
import functools
import os
import time
import yaml
import tqdm
import random
from datetime import datetime
from diffusers import AutoencoderKL
from collections import defaultdict
from dataclasses import dataclass, field
from pathlib import Path
from threading import Lock
from typing import Any, DefaultDict, Dict, List, Literal, Optional, Tuple, Type, cast
import torch
from rich import box, style
from rich.panel import Panel
from rich.table import Table
from torch.cuda.amp.grad_scaler import GradScaler
from prodict import Prodict
import matplotlib.pyplot as plt
import wandb
from functools import partial
from tqdm import trange

import numpy as np
from torchmetrics.functional import structural_similarity_index_measure
from torchmetrics.image import PeakSignalNoiseRatio
from torchmetrics.image.lpip import LearnedPerceptualImagePatchSimilarity

# nerfstudio imports
from nerfstudio.configs.base_config import InstantiateConfig, LoggingConfig, MachineConfig, ViewerConfig
from nerfstudio.engine.callbacks import TrainingCallback, TrainingCallbackAttributes, TrainingCallbackLocation
from nerfstudio.engine.optimizers import OptimizerConfig, Optimizers
from nerfstudio.engine.schedulers import SchedulerConfig, Scheduler
from nerfstudio.utils import profiler
from nerfstudio.utils.decorators import check_eval_enabled, check_main_thread, check_viewer_enabled
from nerfstudio.utils.misc import step_check
from nerfstudio.utils.rich_utils import CONSOLE
from nerfstudio.configs.config_utils import to_immutable_dict
# our imports
from custom_code.utils import TanhNormalizer, get_img_sizes, get_rgb_img_key, step_in_range, interactive_plot
from custom_code.utils import set_requires_grad, optimizer_to
from entrypoint import YAML_CONFIG, BACKGROUND_COLOR_OVERRIDE
from custom_code.custom_method_configs import get_custom_method_configs

TRAIN_INTERATION_OUTPUT = Tuple[torch.Tensor, Dict[str, torch.Tensor], Dict[str, torch.Tensor]]
TORCH_DEVICE = str

ALL_METRICS_TO_LOG = ['metrics_e', 'metrics_d', 'metrics_ae', 'metrics_d-ae']


def range_step_check(step_start, step_end, step_size, run_at_zero=False) -> bool:
    """Returns true based on current step and step interval."""
    if step_size == 0:
        return False
    is_in_range = np.any([step % step_size == 0 for step in range(step_start, step_end)])
    return (run_at_zero or step_start != 0) and is_in_range


def get_metrics_to_log(step, yaml_config):
    loss_e_status = step_in_range(step,  yaml_config.optim.loss_e_range)
    loss_d_status = step_in_range(step,  yaml_config.optim.loss_d_range)
    loss_ae_status = step_in_range(step,  yaml_config.optim.loss_ae_range)
    metrics_to_log = []
    if loss_e_status:
        metrics_to_log.append('metrics_e')
    if loss_d_status:
        metrics_to_log.append('metrics_d')
    if loss_ae_status:
        metrics_to_log.append('metrics_ae')
    return metrics_to_log

class custom_dict(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.counter = 1

    def __missing__(self, key):
        self[key] = custom_dict()
        return self[key]

    def add(self, cd):
        assert set(self.keys()) == set(cd.keys())
        for key in cd.keys():
            self[key] += cd[key]
        self.counter += 1

    def get_avg(self):
        self.scale(1/self.counter)
        return self

    def scale(self, scalar):
        for key in self.keys():
            if isinstance(self[key], torch.Tensor):
                self[key] = self[key].float()
            self[key] *= scalar
        return self

def contruct_logging_metrics_dict(all_scenes_metrics_dict, to_log, metrics_to_log, mode='train'): 
    
    if mode == 'train':
        panel = 'average_single_view_train_metrics'
    elif mode == 'eval':
        panel = 'average_all_view_eval_metrics'
    else:
        raise ValueError('mode not supported')
    
    for metric_type in metrics_to_log:
        dict_avg = None
        for method_name, method_dict in all_scenes_metrics_dict.items():
            dict_method_avg = None
            for method_id, metrics_dict in method_dict.items():
                to_log[f"{method_name}/{method_id}"][f"{mode}_metrics"][metric_type] = metrics_dict[metric_type]
        
                if not dict_method_avg: 
                    dict_method_avg = custom_dict(metrics_dict[metric_type])
                else: 
                    dict_method_avg.add(metrics_dict[metric_type])

            method_avg_metrics = dict_method_avg.get_avg()
            to_log[f"{panel}/{method_name}/{metric_type}"] = method_avg_metrics

            if not dict_avg: 
                dict_avg = custom_dict(method_avg_metrics)
            else: 
                dict_avg.add(method_avg_metrics)

        to_log[f"{panel}/{'all'}/{metric_type}"] = dict_avg.get_avg()

    return to_log


def contruct_logging_losses_dict(all_scenes_loss_dict, to_log): 
    for method_name, method_dict in all_scenes_loss_dict.items():
        dict_method_avg = None
        for method_id, loss_dict in method_dict.items():
            to_log[f"{method_name}/{method_id}"]["train_metrics"] = loss_dict
    
            if not dict_method_avg: 
                dict_method_avg = custom_dict(loss_dict)
            else: 
                dict_method_avg.add(loss_dict)

            method_avg_metrics = dict_method_avg.get_avg()
            to_log[f"average_single_view_train_losses/{method_name}/"] = method_avg_metrics

    return to_log


def recursive_transform_loss_dict(loss_dict):
    logged_dict = {}
    for key, value in loss_dict.items():
        if isinstance(value, dict):
            logged_dict[key] = recursive_transform_loss_dict(value)
        else:
            if 'rgb' in key:
                logged_dict[key.replace('rgb', 'latent')] = value
            else:
                logged_dict[key] = value
    return logged_dict

def free_dict(input_dict: dict):
    "recursive detach of tensors inside input_dict and moves them to cpu"
    res = {}
    for k, v in input_dict.items():
        if isinstance(v, torch.Tensor):
            res[k] = v.detach().cpu()
        elif isinstance(v, dict):
            res[k] = free_dict(v)
    return res

@dataclass
class ExperimentConfig(InstantiateConfig):
    """Full config contents for running an experiment. Any experiment types (like training) will be
    subclassed from this, and must have their _target field defined accordingly."""

    yaml_config: Prodict = None
    """YAML config object"""
    output_dir: Path = Path("outputs")
    """relative or absolute output directory to save all checkpoints and logging"""
    
    experiment_name: Optional[str] = None
    """Experiment name. If None, will automatically be set to dataset name"""
    project_name: Optional[str] = "nerfstudio-project"
    """Project name."""
    timestamp: str = "{timestamp}"
    """Experiment timestamp."""
    logging: LoggingConfig = field(default_factory=LoggingConfig)
    """Logging configuration"""


    vis: Literal["wandb", "viewer+wandb"] = "wandb"
    """Which visualizer to use."""
    relative_model_dir: Path = Path("nerfstudio_models/")
    """Relative path to save all checkpoints."""

    def set_timestamp(self) -> None:
        """Dynamically set the experiment timestamp"""
        if self.timestamp == "{timestamp}":
            self.timestamp = datetime.now().strftime("%Y-%m-%d_%H%M%S")

    def set_experiment_name(self) -> None:
        """Dynamically set the experiment name"""
        if self.experiment_name is None:
            datapath = self.pipeline.datamanager.data
            if datapath is not None:
                datapath = datapath.parent if datapath.is_file() else datapath
                self.experiment_name = str(datapath.stem)
            else:
                self.experiment_name = "unnamed"

    def get_base_dir(self) -> Path:
        """Retrieve the base directory to set relative paths"""
        # check the experiment and method names
        assert self.method_name is not None, "Please set method name in config or via the cli"
        self.set_experiment_name()
        return Path(f"{self.output_dir}/{self.experiment_name}/{self.method_name}/{self.timestamp}")

    def get_checkpoint_dir(self) -> Path:
        """Retrieve the checkpoint directory"""
        return Path(self.get_base_dir() / self.relative_model_dir)

    def print_to_terminal(self) -> None:
        """Helper to pretty print config to terminal"""
        CONSOLE.rule("Config")
        CONSOLE.print(self)
        CONSOLE.rule("")

    def save_config(self) -> None:
        """Save config to base directory"""
        base_dir = self.get_base_dir()
        assert base_dir is not None
        base_dir.mkdir(parents=True, exist_ok=True)
        config_yaml_path = base_dir / "config.yml"
        CONSOLE.log(f"Saving config to: {config_yaml_path}")
        config_yaml_path.write_text(yaml.dump(self), "utf8")

@dataclass
class MultiSceneTrainerConfig(ExperimentConfig):

    _target: Type = field(default_factory=lambda: MultiSceneTrainer)
    """Target for multiscene trainer"""
    methods_and_scenes_map: Dict[str, List[Path]] = field(default_factory=lambda: Dict)
    """Dictionary of method names and their corresponding scene paths."""
    method_name : str = "multiscene_method"
    """constructed method name""" 
    dataset_path: Path = Path("path/to/dataset")
    """Path to dataset."""
    gvae_checkpoint: Optional[Path] = None
    """Path to GVAE checkpoint."""
    steps_per_save: int = 1000
    """Number of steps between saves."""
    steps_per_eval_batch: int = 500
    """Number of steps between randomly sampled batches of rays."""
    steps_per_eval_image: int = 500
    """Number of steps between single eval images."""
    steps_per_eval_all_images: int = 25000
    """Number of steps between eval all images."""
    num_steps: int = 1000000
    """Maximum number of iterations to run."""
    next_scene_batch_every: int = 10
    """Number of mini_steps per step."""

    save_only_latest_checkpoint: bool = True
    """Whether to only save the latest checkpoint or all checkpoints."""
    # optional parameters if we want to resume training
    load_dir: Optional[Path] = None
    """Optionally specify a pre-trained model directory to load from."""
    load_step: Optional[int] = None
    """Optionally specify model step to load from; if none, will find most recent model in load_dir."""
    load_config: Optional[Path] = None
    """Path to config YAML file."""
    load_checkpoint: Optional[Path] = None
    """Path to checkpoint file."""
    log_gradients: bool = False
    """Optionally log gradients during training"""

class MultiSceneTrainer:

    def __init__(self, config: MultiSceneTrainerConfig, local_rank: int = 0) -> None:
        self.config = config
        self.local_rank = local_rank
        self.world_size = 1
        self.device: TORCH_DEVICE = config.device

        if self.device == "cuda":
            self.device += f":{local_rank}"

        if self.device == "cpu":
            self.mixed_precision = False
            CONSOLE.print("Mixed precision is disabled for CPU training.")
        self._start_step: int = 0
        self.base_dir: Path = config.get_base_dir()

        # directory to save checkpoints
        self.checkpoint_dir: Path = config.get_checkpoint_dir()
        CONSOLE.log(f"Saving checkpoints to: {self.checkpoint_dir}")
        self.viewer_state = None
        
        # our code 
        self.img_size, self.render_img_size = get_img_sizes(YAML_CONFIG)
        self.normalizer = TanhNormalizer(scale=YAML_CONFIG.vae.tanh_scale, eps=YAML_CONFIG.vae.eps)

        self.vae = AutoencoderKL.from_pretrained(
            YAML_CONFIG.vae.pretrained_model_name_or_path, 
            **YAML_CONFIG.vae.instantiation_kwargs
        )

        if self.config.gvae_checkpoint:
            checkpoint = torch.load(self.config.gvae_checkpoint, map_location=torch.device('cpu'))
            self.vae.load_state_dict(checkpoint['vae'])
        self.vae.to(self.device)

        psnr_func = PeakSignalNoiseRatio(data_range=(0.0, 1.0))
        ssim_func = partial(structural_similarity_index_measure, data_range=(0.0, 1.0))
        lpips_func = LearnedPerceptualImagePatchSimilarity(normalize=True)
        mse_func = torch.nn.functional.mse_loss
        
        lpips_func.to(self.device)
        psnr_func.to(self.device)
        self.eval_funcs = {
            'psnr_func': psnr_func,
            'ssim_func': ssim_func,
            'lpips_func': lpips_func,
            'mse_func': mse_func
        }
        
    def setup(self, test_mode: Literal["test", "val", "inference"] = "val") -> None:
        """Setup the Trainer by calling other setup functions.

        Args:
            test_mode:
                'val': loads train/val datasets into memory
                'test': loads train/test datasets into memory
                'inference': does not load any dataset into memory
        """

        custom_method_configs = get_custom_method_configs(YAML_CONFIG)

        self.methods = []
        scene_idx = 0
        n_instantiations = sum([len(scene_paths) for scene_paths in self.config.methods_and_scenes_map.values()])
        with tqdm.tqdm(total=n_instantiations, desc="Instantiating methods") as pbar:
            for method_name, scene_paths in self.config.methods_and_scenes_map.items():
                for method_idx, scene_path in enumerate(scene_paths):
                    method_config = custom_method_configs[method_name]
                    method_config.pipeline.datamanager.data = scene_path
                    method_config.pipeline.model.background_color = BACKGROUND_COLOR_OVERRIDE
                    method_config.method_idx = method_idx
                    method_config.scene_idx = scene_idx
                    method = method_config.setup(
                        vae=self.vae,
                        normalizer=self.normalizer,
                        img_size=self.img_size,
                        render_img_size=self.render_img_size,
                        eval_funcs=self.eval_funcs,
                        device=self.device,
                        test_mode=test_mode,
                        local_rank=self.local_rank,
                    )
                    
                    if YAML_CONFIG.multiscene_trainer.loading_strategy == 'gpu_only':
                        method.pipeline.model.to(self.device)

                    self.methods.append(method)
                    scene_idx += 1
                    pbar.update(1)

        # Training is disabled at init
        self.setup_encoder_optim()
        self.setup_decoder_optim()
        #self.setup_nerf_optim() # this is handled by the methods. (required bc of callbacks)
        self.disable_encoder_training()
        self.disable_decoder_training()
        self.disable_nerf_training() # this is handled by the methods

        self._load_checkpoint()

    def setup_encoder_optim(self):
        self.optimizer_encoder = torch.optim.Adam([{
            'params' : self.vae.encoder.parameters(), 
            'lr' : YAML_CONFIG.optim.encoder.lr
        }])

        if YAML_CONFIG.optim.encoder.scheduler.type == 'multistep':
            self.scheduler_encoder = torch.optim.lr_scheduler.MultiStepLR(
                self.optimizer_encoder, 
                milestones=YAML_CONFIG.optim.encoder.scheduler.multistep_config.milestones,
                gamma=YAML_CONFIG.optim.encoder.scheduler.multistep_config.gamma
            )
        elif YAML_CONFIG.optim.encoder.scheduler.type == 'exp':
            self.scheduler_encoder = torch.optim.lr_scheduler.ExponentialLR(
                self.optimizer_encoder, 
                gamma=YAML_CONFIG.optim.encoder.scheduler.exp_config.gamma
            )
        else: 
            raise ValueError('Scheduler type not supported')
    
    def setup_decoder_optim(self):
        self.optimizer_decoder = torch.optim.Adam([{
            'params' : self.vae.decoder.parameters(), 
            'lr' : YAML_CONFIG.optim.decoder.lr
        }])
        if YAML_CONFIG.optim.decoder.scheduler.type == 'multistep':
            self.scheduler_decoder = torch.optim.lr_scheduler.MultiStepLR(
                self.optimizer_decoder, 
                milestones=YAML_CONFIG.optim.decoder.scheduler.multistep_config.milestones,
                gamma=YAML_CONFIG.optim.decoder.scheduler.multistep_config.gamma
            )
        elif YAML_CONFIG.optim.decoder.scheduler.type == 'exp':
            self.scheduler_decoder = torch.optim.lr_scheduler.ExponentialLR(
                self.optimizer_decoder, 
                gamma=YAML_CONFIG.optim.decoder.scheduler.exp_config.gamma
            )
        else: 
            raise ValueError('Scheduler type not supported')
        
    def enable_encoder_training(self): 
        self.train_encoder = True
        set_requires_grad(self.vae.encoder, requires_grad=True)

    def disable_encoder_training(self):
        self.train_encoder = False
        #self.optimizer_encoder = None
        #self.scheduler_encoder = None
        set_requires_grad(self.vae.encoder, requires_grad=False)

    def enable_decoder_training(self): 
        self.train_decoder = True
        set_requires_grad(self.vae.decoder, requires_grad=True)

    def disable_decoder_training(self):
        self.train_decoder = False
        #self.optimizer_decoder = None
        #self.scheduler_decoder = None
        set_requires_grad(self.vae.decoder, requires_grad=False)

    def setup_nerf_optim(self):
        for method in self.methods:
            method.setup_nerf_optim()

    def reset_nerf_optimizers_states(self):
        for method in self.methods:
            method.reset_nerf_optimizers_states()

    def enable_nerf_training(self) -> None:
        self.train_nerf = True
        for method in self.methods:
            method.enable_nerf_training()

    def disable_nerf_training(self) -> None: 
        self.train_nerf = False
        for method in self.methods:
            method.disable_nerf_training()

    def toogle_modules_in_range(self, step:int) ->None: 
        "Enable / disable training of encoder / decoder / nerf modules based on the current step."
        if step_in_range(step, YAML_CONFIG.optim.train_encoder_range):
            if not self.train_encoder: 
                self.enable_encoder_training()
                if self.train_nerf:
                    self.reset_nerf_optimizers_states()
        else:
            if self.train_encoder:
                self.disable_encoder_training()

        if step_in_range(step, YAML_CONFIG.optim.train_decoder_range):
            if not self.train_decoder:
                self.enable_decoder_training()
                if self.train_nerf:
                    self.reset_nerf_optimizers_states()
        else:
            if self.train_decoder:
                self.disable_decoder_training()
                
        if step_in_range(step, YAML_CONFIG.optim.train_nerf_range):
            if not self.train_nerf:
                self.reset_nerf_optimizers_states()
                self.enable_nerf_training()
        else:
            if self.train_nerf:
                self.disable_nerf_training()

    def train(self) -> None:
        """Train the model."""
        
        for method in self.methods:
            assert method.pipeline.datamanager.train_dataset is not None, "Missing DatsetInputs"
            method.pipeline.datamanager.train_dataparser_outputs.save_dataparser_transform(
                self.base_dir / method.method_id / "dataparser_transforms.json"
            )

        num_steps = self.config.num_steps

        torch.cuda.empty_cache()
        for step_start in trange(self._start_step, self._start_step + num_steps, self.config.next_scene_batch_every, desc="Training"):
            step_end = step_start + self.config.next_scene_batch_every

            # eval all methods
            self.eval_iteration(step_start, step_end)

            # train iteration
            for scene_idx in range(len(self.methods)):
                self.methods[scene_idx].pipeline.train()

            all_scenes_loss, all_scenes_loss_dict, all_scenes_metrics_dict = self.train_iterations(step_start=step_start, step_end=step_end)
            
            for k, step in enumerate(range(step_start, step_end)):

                if step_check(step, self.config.logging.steps_per_log, run_at_zero=True):
                    loss_e_status = step_in_range(step,  YAML_CONFIG.optim.loss_e_range)
                    loss_d_status = step_in_range(step,  YAML_CONFIG.optim.loss_d_range)
                    loss_ae_status = step_in_range(step,  YAML_CONFIG.optim.loss_ae_range)

                    to_log = custom_dict({ 
                            'global_information/training_step': step,
                            'global_information/GPU memory (MB)': torch.cuda.max_memory_allocated() / (1024**2),
                            'global_information/all_scenes_loss': all_scenes_loss[k],
                            'training_status/train_nerf': int(self.train_nerf),
                            'training_status/train_encoder': int(self.train_encoder),
                            'training_status/train_decoder': int(self.train_decoder),
                            'training_status/loss_e': int(loss_e_status),
                            'training_status/loss_d': int(loss_d_status),
                            'training_status/loss_ae': int(loss_ae_status),
                        })
                    
                    #     'all_scenes_loss_dict': recursive_transform_loss_dict(all_scenes_loss_dict), # loss details per scene 

                    # Computation of average single-view train metrics and Scene specific single-view train metrics

                    metrics_to_log = get_metrics_to_log(step, YAML_CONFIG)
                    to_log = contruct_logging_losses_dict(all_scenes_loss_dict[k], to_log)
                    to_log = contruct_logging_metrics_dict(all_scenes_metrics_dict[k], to_log, metrics_to_log, mode='train')

                    wandb.log(to_log)

                if step_check(step, self.config.steps_per_save):
                    self.save_checkpoint(step)
                

        # save checkpoint at the end of training
        self.save_checkpoint(step)

        table = Table(
            title=None,
            show_header=False,
            box=box.MINIMAL,
            title_style=style.Style(bold=True),
        )
        table.add_row("Config File", str(self.config.get_base_dir() / "config.yml"))
        table.add_row("Checkpoint Directory", str(self.checkpoint_dir))
        CONSOLE.print(Panel(table, title="[bold][green]:tada: Training Finished :tada:[/bold]", expand=False))

        # after train end callbacks
        for method in self.methods:
            for callback in method.callbacks:
                callback.run_callback_at_location(step=step, location=TrainingCallbackLocation.AFTER_TRAIN)

    def inner_train_iteration(self, scene_idxs, mini_step):
        """Makes one forward pass on each scene  of the batch then the backward and optimization on the total loss"""
        batch_loss_dict = custom_dict()
        batch_metrics_dict = custom_dict()
        batch_loss = 0
        for scene_idx in scene_idxs:
            
            method = self.methods[scene_idx]
            method.callback_before(mini_step)
            loss, loss_dict, metrics_dict = method.forward_pass(mini_step)
            scaled_loss = method.grad_scaler.scale(loss)
            batch_loss += scaled_loss

            # for logging need to detach the tensors in loss_dict and metrics_dict
            batch_loss_dict[method.method_name][method.method_id] = free_dict(loss_dict)
            batch_metrics_dict[method.method_name][method.method_id] = free_dict(metrics_dict)


        # backward pass
        batch_loss.backward()
        
        # optimizer step on nerfs
        for scene_idx in scene_idxs:
            self.methods[scene_idx].optimizer_scheduler_step(mini_step)
            self.methods[scene_idx].callback_after(mini_step)

        # steps on encoder and decoder optimizer / scheduler
        if self.train_encoder:
            self.optimizer_encoder.step()
            self.scheduler_encoder.step()
            self.optimizer_encoder.zero_grad()
        if self.train_decoder:
            self.optimizer_decoder.step()
            self.scheduler_decoder.step()
            self.optimizer_decoder.zero_grad()

        for scene_idx in scene_idxs:
            #self.methods[scene_idx].pipeline.model.zero_grad(set_to_none=True)
            for param_group in self.methods[scene_idx].pipeline.model.get_param_groups().values():
                for param in param_group:
                    param.grad = None

        return batch_loss.detach().cpu(), batch_loss_dict, batch_metrics_dict

    def train_iterations(self, step_start: int, step_end: int) :
        """Calls inner_train_iteration on batches of scenes until all scenes have been updated n_optim_iter times"""
        scene_idx_to_train = list(range(len(self.methods)))
        random.shuffle(scene_idx_to_train)

        all_scenes_loss = [0 for _ in range(step_start, step_end)]
        all_scenes_loss_dict = [{} for _ in range(step_start, step_end)]
        all_scenes_metrics_dict = [{} for _ in range(step_start, step_end)]

        self.toogle_modules_in_range(step=step_start)

        bs = YAML_CONFIG.multiscene_trainer.batch_size
        num_batch = math.ceil(len(scene_idx_to_train) / bs)
        for k in range(0, num_batch):
            scene_idxs = scene_idx_to_train[k*bs: (k+1)*bs]

            # Load on GPU
            if YAML_CONFIG.multiscene_trainer.loading_strategy == 'load_unload':
                self.models_to_device(scene_idxs, self.device)

            # Perform next_scene_batch_every backward pass with the loaded models
            for k, step in enumerate(range(step_start, step_end)):
                batch_loss, batch_loss_dict, batch_metrics_dict = self.inner_train_iteration(scene_idxs, step)
                all_scenes_loss[k] += batch_loss
                all_scenes_loss_dict[k].update(batch_loss_dict)
                all_scenes_metrics_dict[k].update(batch_metrics_dict)

            # Unload from GPU
            if YAML_CONFIG.multiscene_trainer.loading_strategy == 'load_unload':
                self.models_to_device(scene_idxs, 'cpu')

        for k, step in enumerate(range(step_start, step_end)): 
            all_scenes_loss[k] = all_scenes_loss[k] / num_batch

        return all_scenes_loss, all_scenes_loss_dict, all_scenes_metrics_dict
    
    @torch.no_grad()
    def eval_iteration(self, step_start:int, step_end: int) -> None:
        """Run one iteration with different batch/image/all image evaluations depending on step size.

        Args:
            step: Current training step.
        """
        # loggging images
        if range_step_check(step_start, step_end, self.config.steps_per_eval_image):
            to_log = custom_dict({"global_information/training_step": step_start})
            
            for scene_idx, method in enumerate(self.methods): 

                # Load on GPU
                if YAML_CONFIG.multiscene_trainer.loading_strategy == 'load_unload':
                    self.models_to_device([scene_idx], self.device)

                to_log[f"{method.method_name}/figures/{method.method_id}"]["fig_eval"] = method.pipeline.get_images(step=step_start, split='eval')
                to_log[f"{method.method_name}/figures/{method.method_id}"]["fig_train"] = method.pipeline.get_images(step=step_start, split='train')

                # Unload from GPU
                if YAML_CONFIG.multiscene_trainer.loading_strategy == 'load_unload':
                    self.models_to_device([scene_idx], 'cpu')

            plt.close("all")

            wandb.log(to_log)

        # logging psnr across all test images
        if range_step_check(step_start, step_end, self.config.steps_per_eval_all_images):
            """This logs, for train and test views:
                - loss E, loss D, loss AE #TODO
                - nerf_metrics (PSNR, SSIM, LPIPS)
                - decoded_nerf_metrics (PSNR, SSIM, LPIPS)
                - ae_metrics (PSNR, SSIM, LPIPS)
            """
            to_log = custom_dict({"global_information/training_step": step_start})
            all_scenes_metrics_dict = custom_dict()
            
            for scene_idx, method in enumerate(self.methods): 
                pipeline = method.pipeline
                method_id = method.method_id
                method_name = method.method_name

                # Load on GPU
                if YAML_CONFIG.multiscene_trainer.loading_strategy == 'load_unload':
                    self.models_to_device([scene_idx], self.device)

                metrics_dict, metrics_ae_dict, metrics_d_dict, metrics_d_ae_dict = pipeline.get_average_eval_metrics(step=step_start)

                # Unload from GPU
                if YAML_CONFIG.multiscene_trainer.loading_strategy == 'load_unload':
                    self.models_to_device([scene_idx], 'cpu')

                all_scenes_metrics_dict[method_name][method_id]['metrics_e'] = metrics_dict
                all_scenes_metrics_dict[method_name][method_id]['metrics_ae'] = metrics_ae_dict
                all_scenes_metrics_dict[method_name][method_id]['metrics_d'] = metrics_d_dict
                all_scenes_metrics_dict[method_name][method_id]['metrics_d-ae'] = metrics_d_ae_dict

            to_log = contruct_logging_metrics_dict(all_scenes_metrics_dict, to_log, ALL_METRICS_TO_LOG, mode='eval')

            wandb.log(to_log)

    def _load_checkpoint(self) -> None:
        # TODO: Implement this for multiscene
        """Helper function to load pipeline and optimizer from prespecified checkpoint"""
        load_dir = self.config.load_dir
        load_checkpoint = self.config.load_checkpoint
        if load_dir is not None:
            load_step = self.config.load_step
            if load_step is None:
                print("Loading latest Nerfstudio checkpoint from load_dir...")
                # NOTE: this is specific to the checkpoint name format
                load_step = sorted(int(x[x.find("-") + 1 : x.find(".")]) for x in os.listdir(load_dir))[-1]
            load_path: Path = load_dir / f"step-{load_step:09d}.ckpt"
            assert load_path.exists(), f"Checkpoint {load_path} does not exist"
            loaded_state = torch.load(load_path, map_location="cpu")

            CONSOLE.print(f"Loading Nerfstudio checkpoint from {load_path}")

        elif load_checkpoint is not None:
            assert load_checkpoint.exists(), f"Checkpoint {load_checkpoint} does not exist"
            loaded_state = torch.load(load_checkpoint, map_location="cpu")

            CONSOLE.print(f"Loading Nerfstudio checkpoint from {load_checkpoint}")
        else:
            loaded_state = None
            CONSOLE.print("No Nerfstudio checkpoint to load, so training from scratch.")
        
        if loaded_state:
            self._start_step = loaded_state["step"] + 1
            self.vae.load_state_dict(loaded_state['vae'])
            self.optimizer_encoder.load_state_dict(loaded_state["optimizer_encoder"])
            self.scheduler_encoder.load_state_dict(loaded_state["scheduler_encoder"])
            self.optimizer_decoder.load_state_dict(loaded_state["optimizer_decoder"])
            self.scheduler_decoder.load_state_dict(loaded_state["scheduler_decoder"])

            for method in self.methods:
                method.load_state_dict(loaded_state['methods'][method.method_id], loaded_state['step'])

            # load the checkpoints for pipeline, optimizers, and gradient scalar

    @check_main_thread
    def save_checkpoint(self, step: int) -> None:
        """Save the model and optimizers

        Args:
            step: number of steps in training for given checkpoint
        """

        if not self.checkpoint_dir.exists():
            self.checkpoint_dir.mkdir(parents=True, exist_ok=True)
        # save the checkpoint
        ckpt_path: Path = self.checkpoint_dir / f"step-{step:09d}.ckpt"

        # Save the VAE, the optimizers, and the methods
        state = {
            "step": step,
            'vae' : self.vae.state_dict(),
            'optimizer_encoder' : self.optimizer_encoder.state_dict(),
            'scheduler_encoder' : self.scheduler_encoder.state_dict(),
            'optimizer_decoder' : self.optimizer_decoder.state_dict(),
            'scheduler_decoder' : self.scheduler_decoder.state_dict(),
            'methods' : {},
        }
        for method in self.methods:
            state['methods'][method.method_id] = {
                "pipeline": method.pipeline.module.state_dict()  # type: ignore
                if hasattr(method.pipeline, "module")
                else method.pipeline.state_dict(),
                "optimizers": {k: v.state_dict() for (k, v) in method.optimizers.optimizers.items()},
                "schedulers": {k: v.state_dict() for (k, v) in method.optimizers.schedulers.items()},
                "scalers": method.grad_scaler.state_dict(),
            }
        
        torch.save(
            state,
            ckpt_path,
        )

        # possibly delete old checkpoints
        if self.config.save_only_latest_checkpoint:
            # delete everything else in the checkpoint folder
            for f in self.checkpoint_dir.glob("*"):
                if f != ckpt_path:
                    f.unlink()

    def models_to_device(self, scene_idxs, device):
        """Move models and their optimizers specified by `scenes_idxs` to `device`"""
        for scene_idx in scene_idxs:
            method =  self.methods[scene_idx]
            method.pipeline.model.to(device) # validé

            for optimizer in method.optimizers.optimizers.values():
                optimizer_to(optimizer, device)



