import yaml, os
from prodict import Prodict
from custom_code.utils import load_config
from pathlib import Path

def get_user_specific_paths():
    if os.environ['USER'] == 'enter/username/here':
        paths = {
            'data_dir': Path('path/to/data/directory'),
            'gvae_checkpoint_dir': Path('path/to/ig-ae_checkpoints/directory'),
            '3da-ae-v1.5_dir': Path('path/to/3da-ae-v1.5/repository'),
        }
    else: 
        raise NotImplementedError(f"User {os.environ['USER']} not implemented")
    
    return Prodict.from_dict(paths)

VAE_CONFIGS = {
    'runwayml/stable-diffusion-v1-5': {
        'n_latent_channels' : 4,
        'downsample_factor' : 8,
        'bg_color': [0.6176, 0.5711, 0.5046, 0.4422],
        'instantiation_kwargs' : {
            'subfolder' : 'vae',
            'revision' : 'main',
        }
    },
    'ostris/vae-kl-f8-d16' : {
        'n_latent_channels' : 16,
        'downsample_factor' : 8,
        'bg_color': [0.4090, 0.5550, 0.4015, 0.5846, 0.4492, 0.5380, 0.5314, 0.4108, 0.5015,
    0.6012, 0.4245, 0.5039, 0.4920, 0.4955, 0.5296, 0.4999],
        'instantiation_kwargs' : {
            #'torch_dtype' : torch.float16,
        },
    },
}

def get_globals(training_mode, pretrained_model_name_or_path):
    assert pretrained_model_name_or_path in VAE_CONFIGS.keys(), f"Pretrained model {pretrained_model_name_or_path} not supported"
    
    vae_config = VAE_CONFIGS[pretrained_model_name_or_path]
    if training_mode == 'nerfstudio': 
        num_channels = 3
        bg_color_override = [1., 1., 1.]

    elif training_mode == 'multiscene':
        num_channels = vae_config['n_latent_channels']
        bg_color_override = vae_config['bg_color']

    else:
        raise NotImplementedError(f"Training mode {training_mode} not supported")
    
    return training_mode, num_channels, bg_color_override, vae_config

def load_from_env(arg_name, default):
    if arg_name in os.environ.keys():
        return os.environ[arg_name]
    else:
        return default


# ----- Argparse emulation -----
#config = load_from_env('config', 'default.yaml') 
config = load_from_env('config', 'rendering_time/nfs_ngp.yaml')

gvae_checkpoint_override = load_from_env('gvae_checkpoint_override', None)
pretrained_model_name_or_path = load_from_env('pretrained_model_name_or_path', None)
exp_name_suffix = load_from_env('exp_name_suffix', None)
# ----- Argparse emulation end -----

# Load config
repo_path = os.path.dirname(os.path.realpath(__file__))
yaml_config_dir = os.path.join(repo_path, "configs")
yaml_config = Prodict.from_dict(load_config(config, yaml_config_dir, from_default=True, default_cfg_name="default.yaml"))

# Overriding config if necessary
if gvae_checkpoint_override:
    yaml_config.vae.gvae_checkpoint = gvae_checkpoint_override
    print(f"Overriding gvae_checkpoint with '{gvae_checkpoint_override}'")
if pretrained_model_name_or_path:
    yaml_config.vae.pretrained_model_name_or_path = pretrained_model_name_or_path
    print(f"Overriding pretrained_model_name_or_path with '{pretrained_model_name_or_path}'")
if exp_name_suffix:
    yaml_config.experiment_name_prefix = yaml_config.experiment_name
    yaml_config.experiment_name_suffix = exp_name_suffix
    yaml_config.experiment_name = f"{yaml_config.experiment_name}__{exp_name_suffix}"
    print(f"Completing experiment_name with '{exp_name_suffix}'")


# fetching globals
training_mode, num_channels, background_color_override, vae_cfg = get_globals(yaml_config.training_mode, yaml_config.vae.pretrained_model_name_or_path)
paths = get_user_specific_paths()
yaml_config['vae'].update(vae_cfg)
yaml_config['globals'] = {
    'training_mode': training_mode,
    'num_channels': num_channels,
    'background_color_override': background_color_override
}

# defining globals
PATHS = paths
YAML_CONFIG = yaml_config
TRAINING_MODE = training_mode
NUM_CHANNELS = num_channels
BACKGROUND_COLOR_OVERRIDE = background_color_override

    
