import os,random, torch
import numpy as np
from pathlib import Path
from prodict import Prodict
from typing import Optional

# custom imports
from entrypoint import YAML_CONFIG, BACKGROUND_COLOR_OVERRIDE, PATHS
from custom_code.custom_method_configs import get_custom_method_configs, get_dataset_specific_args
from custom_code.utils import MethodsAndScenesMapper

from nerfstudio.utils import comms, profiler
from nerfstudio.configs.method_configs import method_configs as nefstudio_method_configs
from nerfstudio.data.dataparsers.nerfstudio_dataparser import NerfstudioDataParserConfig
from nerfstudio.data.dataparsers.blender_dataparser import BlenderDataParserConfig
import wandb
from nerfstudio.configs.config_utils import to_immutable_dict
from custom_code.multiscene_trainer import MultiSceneTrainer, MultiSceneTrainerConfig

def _set_random_seed(seed) -> None:
    """Set randomness seed in torch and numpy"""
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)

def construct_name(methods_repartition):
    m_name = ""
    for method_name, scene_paths in methods_repartition.items():
        num_scenes = len(scene_paths)
        if num_scenes > 0:
            m_name += f"{num_scenes}{method_name[0:2]}_"
    return m_name[:-1]

def train_loop(config: MultiSceneTrainerConfig, local_rank: int = 0):
    """Main training function that sets up and runs the trainer per process

    Args:
        config: config file specifying training regimen
    """
    trainer = config.setup(local_rank=local_rank)
    trainer.setup()
    trainer.train()
    profiler.flush_profiler(config.logging)

def fix_gvae_checkpoint(gvae_checkpoint:Optional[Path]):
    if gvae_checkpoint:
        if os.path.isabs(gvae_checkpoint):
            return gvae_checkpoint
        
        return PATHS.gvae_checkpoint_dir / gvae_checkpoint


if __name__ == '__main__' :
    # method to data mapping management
    methods_and_scenes_mapper = MethodsAndScenesMapper(PATHS.data_dir, YAML_CONFIG)

    # rgb nerf entrypoint (nerfstudio)
    if YAML_CONFIG.training_mode == 'nerfstudio':

        method_name = YAML_CONFIG.nerfstudio_trainer.method_name
        print(f"Choosing nerfstudio method config for {method_name}")
        chosen_trainer_config = nefstudio_method_configs[method_name]

        dataparser_config, near_plane, far_plane = get_dataset_specific_args(YAML_CONFIG)
        chosen_trainer_config.pipeline.datamanager.dataparser = dataparser_config
        try:
            chosen_trainer_config.pipeline.model.collider_params["near_plane"] = near_plane
            chosen_trainer_config.pipeline.model.collider_params["far_plane"] = far_plane
        except:
            pass # no collider for current method
        finally:
            chosen_trainer_config.pipeline.model.near_plane = near_plane
            chosen_trainer_config.pipeline.model.far_plane = far_plane
        chosen_trainer_config.vis='wandb'
        chosen_trainer_config.pipeline.datamanager.data = methods_and_scenes_mapper.get_scene_path(YAML_CONFIG.nerfstudio_trainer.scene_id_or_idx)
        chosen_trainer_config.pipeline.model.background_color = BACKGROUND_COLOR_OVERRIDE
        chosen_trainer_config.method_name = YAML_CONFIG.nerfstudio_trainer.method_name
        chosen_trainer_config.max_num_iterations = YAML_CONFIG.optim.num_steps
        chosen_trainer_config.steps_per_eval_batch = YAML_CONFIG.logging.steps_per_eval_batch
        chosen_trainer_config.steps_per_eval_image = YAML_CONFIG.logging.steps_per_eval_image
        chosen_trainer_config.steps_per_eval_all_images = YAML_CONFIG.logging.steps_per_eval_all_images
        chosen_trainer_config.steps_per_save = YAML_CONFIG.logging.steps_per_save
        chosen_trainer_config.experiment_name = YAML_CONFIG.experiment_name

    
    # custom method entrypoint
    elif YAML_CONFIG.training_mode == 'multiscene': 
        
        method_name = construct_name(YAML_CONFIG.multiscene_trainer.methods_repartition)
        chosen_trainer_config = MultiSceneTrainerConfig(
            methods_and_scenes_map=methods_and_scenes_mapper.get_methods_and_scenes_map(
                YAML_CONFIG.multiscene_trainer.methods_repartition
            ),
            method_name=method_name,
            experiment_name=YAML_CONFIG.experiment_name,
            gvae_checkpoint=fix_gvae_checkpoint(YAML_CONFIG.vae.gvae_checkpoint),
            steps_per_save=YAML_CONFIG.logging.steps_per_save,
            steps_per_eval_batch=YAML_CONFIG.logging.steps_per_eval_batch,
            steps_per_eval_image = YAML_CONFIG.logging.steps_per_eval_image,
            steps_per_eval_all_images=YAML_CONFIG.logging.steps_per_eval_all_images,
            num_steps=YAML_CONFIG.optim.num_steps,
            next_scene_batch_every=YAML_CONFIG.optim.next_scene_batch_every
        )
    else:
        raise NotImplementedError(f"Training mode {YAML_CONFIG.training_mode} not supported")
    
    chosen_trainer_config.logging.steps_per_log = YAML_CONFIG.logging.steps_per_log
    # chosen_trainer_config.logging.max_buffer_size = YAML_CONFIG.logging.logging_buffer_size
    
    chosen_trainer_config.device = YAML_CONFIG.device
    chosen_trainer_config.project_name = YAML_CONFIG.project_name
    
    chosen_trainer_config.set_timestamp()
    chosen_trainer_config.print_to_terminal()
    chosen_trainer_config.save_config()

    if YAML_CONFIG.training_mode == 'multiscene' :
        # Initializing wandb
        wandb.init(
            project=YAML_CONFIG.project_name,
            config=YAML_CONFIG,
            entity="wandb_project_name",
            name=YAML_CONFIG.experiment_name,
        )


    # launch training
    _set_random_seed(YAML_CONFIG.random_seed)
    train_loop(config=chosen_trainer_config, local_rank=YAML_CONFIG.local_rank)